"""
Test Creator module.

This module contains tests for Creator objects.
"""
from datetime import date, datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import creators_list, exceptions


def test_known_creator(talker):
    """Test for a known creator."""
    jack = talker.creator(3)
    assert jack.name == "Jack Kirby"
    assert jack.birth == date(1917, 8, 28)
    assert jack.death == date(1994, 2, 6)
    assert (
        jack.image
        == "https://static.metron.cloud/media/creator/2018/11/11/432124-Jack_Kirby01.jpg"
    )
    assert jack.wikipedia == "Jack_Kirby"
    assert jack.modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        22,
        311024,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_comiclist(talker):
    """Test the CreatorsList."""
    creators = talker.creators_list({"name": "man"})
    creator_iter = iter(creators)
    assert next(creator_iter).name == "A. J. Lieberman"
    assert next(creator_iter).name == "Adam Freeman"
    assert next(creator_iter).name == "Adam Schlagman"
    assert next(creator_iter).name == "Al Sulman"
    assert len(creators) == 177
    assert creators[3].name == "Al Sulman"


def test_bad_creator(talker):
    """Test for a non-existant creator."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/creator/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.creator(-1)


def test_bad_response_data(talker):
    """Test for a bad creator response."""
    with pytest.raises(exceptions.ApiError):
        creators_list.CreatorsList({"results": {"name": 1}})



"""
Test Arcs module.

This module contains tests for Arc objects.
"""
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import arcs_list, exceptions


def test_known_arc(talker):
    """Test for known arcs."""
    heroes = talker.arc(1)
    assert heroes.name == "Heroes In Crisis"
    assert (
        heroes.image
        == "https://static.metron.cloud/media/arc/2018/11/12/heroes-in-crisis.jpeg"
    )
    assert heroes.modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        19,
        456634,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_arcslist(talker):
    """Test for ArcsList."""
    arcs = talker.arcs_list()
    arc_iter = iter(arcs)
    assert next(arc_iter).name == "2099"
    assert next(arc_iter).name == "52"
    assert next(arc_iter).name == "A Court of Owls"
    assert len(arcs) == 571
    assert arcs[2].name == "A Court of Owls"


def test_bad_arc(talker):
    """Test for bad arc requests."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/arc/-8/",
            text='{"response_code": 404, "detail": "Not found."}',
        )

        with pytest.raises(exceptions.ApiError):
            talker.arc(-8)


def test_bad_response_data():
    """Test for bad arc response."""
    with pytest.raises(exceptions.ApiError):
        arcs_list.ArcsList({"results": {"name": 1}})



"""
CreatorsList module.

This module provides the following classes:

- CreatorsList
"""
from marshmallow import ValidationError

from mokkari import creator, exceptions


class CreatorsList:
    """The CreatorsList object contains a list of `Creator` objects."""

    def __init__(self, response) -> None:
        """Initialize a new CreatorsList."""
        self.creators = []

        schema = creator.CreatorSchema()
        for creator_dict in response["results"]:
            try:
                result = schema.load(creator_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.creators.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.creators)

    def __len__(self):
        """Return the length of the object."""
        return len(self.creators)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.creators[index]



"""
TeamsList module.

This module provides the following classes:

- TeamsList
"""
from marshmallow import ValidationError

from mokkari import exceptions, team


class TeamsList:
    """The TeamsList object contains a list of `Team` objects."""

    def __init__(self, response):
        """Initialize a new TeamsList."""
        self.teams = []

        schema = team.TeamSchema()
        for team_dict in response["results"]:
            try:
                result = schema.load(team_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.teams.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.teams)

    def __len__(self):
        """Return the length of the object."""
        return len(self.teams)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.teams[index]



"""
Variant module.

This module provides the following classes:

- Variant
- VariantSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load


class Variant:
    """
    The Variant object contains information for variant issues.

    :param `**kwargs`: The keyword arguments is used for setting variant issue data.
    """

    def __init__(self, **kwargs):
        """Intialize a new Variant."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class VariantSchema(Schema):
    """Schema for the Variant API."""

    name = fields.Str()
    sku = fields.Str()
    image = fields.Url()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Variant object.

        :param data: Data from Metron response.

        :returns: :class:`Variant` object
        :rtype: Variant
        """
        return Variant(**data)



"""
Test Series module.

This module contains tests for Series objects.
"""
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import exceptions, series_list


def test_known_series(talker):
    """Test for a known series."""
    death = talker.series(1)
    assert death.name == "Death of the Inhumans"
    assert death.sort_name == "Death of the Inhumans"
    assert death.volume == 1
    assert death.year_began == 2018
    assert death.year_end == 2018
    assert death.issue_count == 5
    assert death.image == "https://static.metron.cloud/media/issue/2018/11/11/6497376-01.jpg"
    assert death.series_type.name == "Mini-Series"
    assert death.publisher_id == 1
    assert death.modified == datetime(
        2019,
        7,
        5,
        14,
        32,
        52,
        239629,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_series_without_year_end(talker):
    """Test for series without a year-end date."""
    abs_carnage = talker.series(2311)
    assert abs_carnage.name == "Absolute Carnage"
    assert abs_carnage.sort_name == "Absolute Carnage"
    assert abs_carnage.volume == 1
    assert abs_carnage.year_began == 2019
    assert abs_carnage.year_end is None
    assert abs_carnage.issue_count == 5
    assert abs_carnage.publisher_id == 1
    assert abs_carnage.series_type.name == "Mini-Series"


def test_serieslist(talker):
    """Test the SeriesList."""
    series = talker.series_list({"name": "batman"})
    series_iter = iter(series)
    assert next(series_iter).id == 2547
    assert next(series_iter).id == 2481
    assert next(series_iter).id == 763
    assert next(series_iter).id == 93
    assert len(series) == 115
    assert series[3].id == 93
    assert series[3].display_name == "Batman (2016)"


def test_bad_series(talker):
    """Test for a non-existant series."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/series/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.series(-1)


def test_bad_response_data():
    """Test for a bad series response."""
    with pytest.raises(exceptions.ApiError):
        series_list.SeriesList({"results": {"name": 1}})



"""
Creator module.

This module provides the following classes:

- Creator
- CreatorSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load


class Creator:
    """
    The Creator object contains information for creators.

    :param `**kwargs`: The keyword arguments is used for setting creator data from Metron.
    """

    def __init__(self, **kwargs):
        """Intialize a new Creator."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CreatorSchema(Schema):
    """
    Schema for the Creator API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field
    """

    id = fields.Int()
    name = fields.Str()
    birth = fields.Date(allow_none=True)
    death = fields.Date(allow_none=True)
    desc = fields.Str()
    wikipedia = fields.Str()
    image = fields.Url()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        dateformat = "%Y-%m-%d"
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Creator object.

        :param data: Data from Metron response.

        :returns: :class:`Creator` object
        :rtype: Creator
        """
        return Creator(**data)



"""
Session module.

This module provides the following classes:

- Session
"""
import platform
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlencode

import requests
from marshmallow import ValidationError
from ratelimit import limits, sleep_and_retry

from mokkari import (
    __version__,
    arc,
    arcs_list,
    character,
    characters_list,
    creator,
    creators_list,
    exceptions,
    issue,
    issues_list,
    publisher,
    publishers_list,
    series,
    series_list,
    sqlite_cache,
    team,
    teams_list,
)

ONE_MINUTE = 60


class Session:
    """
    Session to request api endpoints.

    :param str username: The username for authentication with metron.cloud
    :param str passwd: The password used for authentication with metron.cloud
    :param SqliteCache optional: SqliteCache to use
    """

    def __init__(
        self,
        username: str,
        passwd: str,
        cache: Optional[sqlite_cache.SqliteCache] = None,
    ) -> None:
        """Intialize a new Session."""
        self.username = username
        self.passwd = passwd
        self.header = {
            "User-Agent": f"Mokkari/{__version__} ({platform.system()}; {platform.release()})"
        }
        self.api_url = "https://metron.cloud/api/{}/"
        self.cache = cache

    def call(
        self, endpoint: List[Union[str, int]], params: Dict[str, Union[str, int]] = None
    ) -> Dict[str, Any]:
        """
        Make request for api endpoints.

        :param str endpoint: The endpoint to request information from.
        :param dict params: Parameters to add to the request.
        """
        if params is None:
            params = {}

        cache_params = ""
        if params:
            ordered_params = OrderedDict(sorted(params.items(), key=lambda t: t[0]))
            cache_params = "?{}".format(urlencode(ordered_params))

        url = self.api_url.format("/".join(str(e) for e in endpoint))
        cache_key = f"{url}{cache_params}"

        cached_response = self._get_results_from_cache(cache_key)
        if cached_response is not None:
            return cached_response

        data = self._request_data(url, params)

        if "detail" in data:
            raise exceptions.ApiError(data["detail"])

        self._save_results_to_cache(cache_key, data)

        return data

    def creator(self, _id: int) -> creator.Creator:
        """
        Request data for a creator based on its ``_id``.

        :param int _id: The creator id.

        :return: :class:`Creator` object
        :rtype: Creator
        """
        try:
            result = creator.CreatorSchema().load(self.call(["creator", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def creators_list(self, params: Optional[Dict[str, Union[str, int]]] = None):
        """
        Request a list of creators.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Creator` objects containing their id and name.
        :rtype: CreatorsList
        """
        if params is None:
            params = {}

        res = self.call(["creator"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return creators_list.CreatorsList(res)

    def character(self, _id: int) -> character.Character:
        """
        Request data for a character based on its ``_id``.

        :param int _id: The character id.

        :return: :class:`Character` object
        :rtype: Character
        """
        try:
            result = character.CharacterSchema().load(self.call(["character", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def characters_list(self, params: Optional[Dict[str, Union[str, int]]] = None):
        """
        Request a list of characters.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Character` objects containing their id and name.
        :rtype: CharactersList
        """
        if params is None:
            params = {}

        res = self.call(["character"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return characters_list.CharactersList(res)

    def publisher(self, _id: int) -> publisher.Publisher:
        """
        Request data for a publisher based on its ``_id``.

        :param int _id: The publisher id.

        :return: :class:`Publisher` object
        :rtype: Publisher
        """
        try:
            result = publisher.PublisherSchema().load(self.call(["publisher", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def publishers_list(self, params: Dict[str, Union[str, int]] = None):
        """
        Request a list of publishers.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Publisher` objects containing their id and name.
        :rtype: PublishersList
        """
        if params is None:
            params = {}

        res = self.call(["publisher"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return publishers_list.PublishersList(res)

    def team(self, _id: int) -> team.Team:
        """
        Request data for a team based on its ``_id``.

        :param int _id: The team id.

        :return: :class:`Team` object
        :rtype: Team
        """
        try:
            result = team.TeamSchema().load(self.call(["team", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def teams_list(self, params: Dict[str, Union[str, int]] = None):
        """
        Request a list of teams.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Team` objects containing their id and name.
        :rtype: TeamsList
        """
        if params is None:
            params = {}

        res = self.call(["team"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return teams_list.TeamsList(res)

    def arc(self, _id: int) -> arc.Arc:
        """
        Request data for a story arc based on its ``_id``.

        :param int _id: The story arc id.

        :return: :class:`Arc` object
        :rtype: Arc
        """
        try:
            result = arc.ArcSchema().load(self.call(["arc", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def arcs_list(self, params: Dict[str, Union[str, int]] = None):
        """
        Request a list of story arcs.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Arc` objects containing their id and name.
        :rtype: ArcsList
        """
        if params is None:
            params = {}

        res = self.call(["arc"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return arcs_list.ArcsList(res)

    def series(self, _id: int) -> series.Series:
        """
        Request data for a series based on its ``_id``.

        :param int _id: The series id.

        :return: :class:`Series` object
        :rtype: Series
        """
        try:
            result = series.SeriesSchema().load(self.call(["series", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def series_list(self, params: Dict[str, Union[str, int]] = None):
        """
        Request a list of series.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Series` objects containing their id and name.
        :rtype: SeriesList
        """
        if params is None:
            params = {}

        res = self.call(["series"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return series_list.SeriesList(res)

    def issue(self, _id: int) -> issue.Issue:
        """
        Request data for an issue based on it's ``_id``.

        :param int _id: The issue id.

        :return: :class:`Issue` object
        :rtype: Issue
        """
        try:
            result = issue.IssueSchema().load(self.call(["issue", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error)

        return result

    def issues_list(self, params: Dict[str, Union[str, int]] = None):
        """
        Request a list of issues.

        :param params: Parameters to add to the request.
        :type params: dict, optional

        :return: A list of :class:`Issue` objects containing their id and name.
        :rtype: IssuesList
        """
        if params is None:
            params = {}

        res = self.call(["issue"], params=params)
        if res["next"]:
            res = self._retrieve_all_results(res)

        return issues_list.IssuesList(res)

    def _retrieve_all_results(self, data):
        has_next_page = True
        next_page = data["next"]

        while has_next_page:
            cached_response = self._get_results_from_cache(next_page)
            if cached_response:
                data["results"].extend(cached_response["results"])
                if cached_response["next"]:
                    next_page = cached_response["next"]
                else:
                    has_next_page = False
                continue

            response = self._request_data(next_page)
            data["results"].extend(response["results"])

            self._save_results_to_cache(next_page, response)

            if response["next"]:
                next_page = response["next"]
            else:
                has_next_page = False

        return data

    @sleep_and_retry
    @limits(calls=20, period=ONE_MINUTE)
    def _request_data(
        self, url: str, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> Any:
        if params is None:
            params = {}

        try:
            response = requests.get(
                url,
                params=params,
                auth=(self.username, self.passwd),
                headers=self.header,
            ).json()
        except requests.exceptions.ConnectionError as e:
            raise exceptions.ApiError("Connection error: {}".format(repr(e)))

        return response

    def _get_results_from_cache(self, key) -> Optional[Any]:
        cached_response = None

        if self.cache:
            try:
                cached_response = self.cache.get(key)
            except AttributeError as e:
                raise exceptions.CacheError(
                    "Cache object passed in is missing attribute: {}".format(repr(e))
                )

        return cached_response

    def _save_results_to_cache(self, key: str, data: str) -> None:
        if self.cache:
            try:
                self.cache.store(key, data)
            except AttributeError as e:
                raise exceptions.CacheError(
                    "Cache object passed in is missing attribute: {}".format(repr(e))
                )



"""
Test Cache module.

This module contains tests for SqliteCache objects.
"""
import json

import pytest
import requests_mock

from mokkari import api, exceptions, sqlite_cache


class NoGet:
    """The NoGet object fakes storing data from the sqlite cache."""

    def store(self, key, value):
        """Save no data."""
        # This method should store key value pair
        return


class NoStore:
    """The NoStore object fakes getting data from the sqlite cache."""

    def get(self, key):
        """Retrieve no data."""
        return None


def test_no_get(dummy_username, dummy_password):
    """Test for retrieving failure."""
    m = api(username=dummy_username, passwd=dummy_password, cache=NoGet())

    with pytest.raises(exceptions.CacheError):
        m.series(5)


def test_no_store(dummy_username, dummy_password):
    """Test for saving data error."""
    m = api(username=dummy_username, passwd=dummy_password, cache=NoStore())

    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/series/5/",
            text='{"response_code": 200}',
        )

        with pytest.raises(exceptions.CacheError):
            m.series(5)


def test_sql_store(dummy_username, dummy_password):
    """Test for saving data to the sqlite cache."""
    fresh_cache = sqlite_cache.SqliteCache(":memory:")
    test_cache = sqlite_cache.SqliteCache("tests/testing_mock.sqlite")

    m = api(username=dummy_username, passwd=dummy_password, cache=fresh_cache)
    url = "https://metron.cloud/api/series/1/"

    assert fresh_cache.get(url) is None

    try:
        with requests_mock.Mocker() as r:
            r.get(url, text=json.dumps(test_cache.get(url)))
            m.series(1)

        assert fresh_cache.get(url) is not None
    except TypeError:
        print(
            "This test will fail after cache db deleted.\n"
            "It should pass if you now re-run the test suite without deleting the database."
        )
        assert False



"""
Test Publishers module.

This module contains tests for Publisher objects.
"""
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import exceptions, publishers_list


def test_known_publishers(talker):
    """Test for a known publisher."""
    marvel = talker.publisher(1)
    assert marvel.name == "Marvel"
    assert marvel.image == "https://static.metron.cloud/media/publisher/2018/11/11/marvel.jpg"
    assert marvel.wikipedia == "Marvel_Comics"
    assert marvel.founded == 1939
    assert marvel.modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        23,
        591390,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_publisherlist(talker):
    """Test the PublishersList."""
    publishers = talker.publishers_list()
    publisher_iter = iter(publishers)
    assert next(publisher_iter).name == "12-Gauge Comics"
    assert next(publisher_iter).name == "AWA Studios"
    assert next(publisher_iter).name == "AfterShock Comics"
    assert len(publishers) == 44
    assert publishers[2].name == "AfterShock Comics"


def test_bad_publisher(talker):
    """Test for a non-existant publisher."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/publisher/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.publisher(-1)


def test_bad_response_data():
    """Test for a bad publisher response."""
    with pytest.raises(exceptions.ApiError):
        publishers_list.PublishersList({"results": {"name": 1}})



"""
Test Characters module.

This module contains tests for Character objects.
"""
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import characters_list, exceptions


def test_known_character(talker):
    """Test for a known character."""
    black_bolt = talker.character(1)
    assert black_bolt.name == "Black Bolt"
    assert (
        black_bolt.image
        == "https://static.metron.cloud/media/character/2018/11/11/black-bolt.jpg"
    )
    assert black_bolt.wikipedia == "Black_Bolt"
    assert len(black_bolt.creators) == 2
    assert len(black_bolt.teams) == 3
    assert black_bolt.modified == datetime(
        2021,
        9,
        9,
        15,
        52,
        49,
        90281,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_characterlist(talker):
    """Test the CharactersList."""
    character = talker.characters_list({"name": "man"})
    character_iter = iter(character)
    assert next(character_iter).name == "'Mazing Man"
    assert next(character_iter).name == "3-D Man (Chandler)"
    assert next(character_iter).name == "3-D Man (Garrett)"
    assert len(character) == 435
    assert character[2].name == "3-D Man (Garrett)"


def test_bad_character(talker):
    """Test for a non-existing character."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/character/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.character(-1)


def test_bad_response_data(talker):
    """Teset for a bad character response."""
    with pytest.raises(exceptions.ApiError):
        characters_list.CharactersList({"results": {"name": 1}})



"""
Test Teams module.

This module contains tests for Team objects.
"""
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import exceptions, teams_list


def test_known_team(talker):
    """Test for a known team."""
    inhumans = talker.team(1)
    assert inhumans.name == "Inhumans"
    assert inhumans.image == "https://static.metron.cloud/media/team/2018/11/11/Inhumans.jpg"
    assert inhumans.wikipedia == "Inhumans"
    assert len(inhumans.creators) == 2
    assert inhumans.modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        23,
        975156,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_teamlist(talker):
    """Test the TeamsList."""
    teams = talker.teams_list()
    team_iter = iter(teams)
    assert next(team_iter).name == "A-Force"
    assert next(team_iter).name == "A.I.M."
    assert next(team_iter).name == "A.R.G.U.S."
    assert len(teams) == 456
    assert teams[2].name == "A.R.G.U.S."


def test_bad_team(talker):
    """Test for a non-existant team."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/team/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.team(-1)


def test_bad_response_data():
    """Test for a bad team response."""
    with pytest.raises(exceptions.ApiError):
        teams_list.TeamsList({"results": {"name": 1}})



"""
Series module.

This module provides the following classes:

- SeriesType
- SeriesTypeSchema
- Series
- SeriesSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load


class SeriesType:
    """
    The SeriesType object contains information for type of series.

    :param `**kwargs`: The keyword arguments is used for setting series type data.
    """

    def __init__(self, **kwargs):
        """Intialize a new SeriesType."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class SeriesTypeSchema(Schema):
    """Schema for the Series Type."""

    id = fields.Int()
    name = fields.Str()

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the SeriesType object.

        :param data: Data from Metron response.

        :returns: :class:`SeriesType` object
        :rtype: SeriesType
        """
        return SeriesType(**data)


class Series:
    """
    The Series object contains information for comic series.

    :param `**kwargs`: The keyword arguments is used for setting series data.
    """

    def __init__(self, **kwargs):
        """Intialize a new Series."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class SeriesSchema(Schema):
    """
    Schema for the Series API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field
    """

    id = fields.Int()
    name = fields.Str()
    sort_name = fields.Str()
    volume = fields.Int()
    series_type = fields.Nested(SeriesTypeSchema)
    publisher = fields.Int(attribute="publisher_id")
    year_began = fields.Int()
    year_end = fields.Int(allow_none=True)
    desc = fields.Str()
    issue_count = fields.Int()
    image = fields.Url()
    display_name = fields.Str(data_key="__str__")
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Series object.

        :param data: Data from Metron response.

        :returns: :class:`Series` object
        :rtype: Series
        """
        return Series(**data)



"""
Character module.

This module provides the following classes:

- Character
- CharacterSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load

from mokkari import creator, team


class Character:
    """
    The Character object contains information for characters.

    :param `**kwargs`: The keyword arguments is used for setting character data from Metron.
    """

    def __init__(self, **kwargs):
        """Intialize a new Character."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CharacterSchema(Schema):
    """
    Schema for the Character API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field
    """

    id = fields.Int()
    name = fields.Str()
    alias = fields.List(fields.Str)
    desc = fields.Str()
    wikipedia = fields.Str()
    image = fields.Url()
    creators = fields.Nested(creator.CreatorSchema, many=True)
    teams = fields.Nested(team.TeamSchema, many=True)
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the character object.

        :param data: Data from Metron response.

        :returns: :class:`Character` object
        :rtype: Character
        """
        return Character(**data)



"""
ArcsList module.

This module provides the following classes:

- ArcsList
"""
from marshmallow import ValidationError

from mokkari import arc, exceptions


class ArcsList:
    """The ArcsList object contains a list of `Arc` objects."""

    def __init__(self, response):
        """Initialize a new ArcsList."""
        self.arcs = []

        schema = arc.ArcSchema()
        for arc_dict in response["results"]:
            try:
                result = schema.load(arc_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.arcs.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.arcs)

    def __len__(self):
        """Return the length of the object."""
        return len(self.arcs)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.arcs[index]



"""Project entry file."""

# Keep this at beginning of file to prevent circular import with session
__version__ = "1.0.2"

from typing import Optional

from mokkari import exceptions, session, sqlite_cache


def api(
    username: Optional[str] = None,
    passwd: Optional[str] = None,
    cache: sqlite_cache.SqliteCache = None,
):
    """Entry function the sets login credentials for metron.cloud.

    :param str username: The username used for metron.cloud.
    :param str passwd: The password used for metron.cloud.
    :param SqliteCache optional: SqliteCache to use
    """
    if username is None:
        raise exceptions.AuthenticationError("Missing username.")

    if passwd is None:
        raise exceptions.AuthenticationError("Missing passwd.")

    return session.Session(username, passwd, cache=cache)



"""
Arc module.

This module provides the following classes:

- Arc
- ArcSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load


class Arc:
    """
    The Arc object contains information for story arcs.

    :param `**kwargs`: The keyword arguments is used for setting arc data from Metron.
    """

    def __init__(self, **kwargs):
        """Intialize a new Arc."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class ArcSchema(Schema):
    """
    Schema for the Arc API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field
    """

    id = fields.Int()
    name = fields.Str()
    desc = fields.Str()
    image = fields.Url()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the arc object.

        :param data: Data from Metron response.

        :returns: :class:`Arc` object
        :rtype: Arc
        """
        return Arc(**data)



"""
SeriesList module.

This module provides the following classes:

- SeriesList
"""
from marshmallow import ValidationError

from mokkari import exceptions, series


class SeriesList:
    """The SeriesList object contains a list of `Series` objects."""

    def __init__(self, response):
        """Initialize a new SeriesList."""
        self.series = []

        schema = series.SeriesSchema()
        for series_dict in response["results"]:
            try:
                result = schema.load(series_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.series.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.series)

    def __len__(self):
        """Return the length of the object."""
        return len(self.series)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.series[index]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mokkari']

package_data = \
{'': ['*']}

install_requires = \
['marshmallow>=3.13.0,<4.0.0',
 'ratelimit>=2.2.1,<3.0.0',
 'requests>=2.26.0,<3.0.0']

extras_require = \
{'docs': ['sphinxcontrib-napoleon>=0.7,<0.8', 'sphinx-rtd-theme>=0.5.2,<0.6.0']}

setup_kwargs = {
    'name': 'mokkari',
    'version': '1.0.2',
    'description': 'Python wrapper for Metron API',
    'long_description': '=======\nMokkari\n=======\n.. image:: https://img.shields.io/pypi/v/mokkari.svg?logo=PyPI&label=Version&style=flat-square   :alt: PyPI\n    :target: https://pypi.org/project/mokkari\n\n.. image:: https://img.shields.io/pypi/pyversions/mokkari.svg?logo=Python&label=Python-Versions&style=flat-square\n    :target: https://pypi.org/project/mokkari\n\n.. image:: https://img.shields.io/github/license/bpepple/mokkari\n    :target: https://opensource.org/licenses/GPL-3.0  \n\n.. image:: https://img.shields.io/badge/Code%20Style-Black-000000.svg?style=flat-square\n    :target: https://github.com/psf/black\n\nQuick Description\n-----------------\nA python wrapper for the metron.cloud_ API.\n\n.. _metron.cloud: https://metron.cloud\n\nInstallation\n------------\n\nPyPi\n~~~~\n\n.. code:: bash\n\n  $ pip3 install --user mokkari\n\nExample Usage\n-------------\n.. code-block:: python\n\n    import mokkari\n\n    # Your own config file to keep your credentials secret\n    from config import username, password\n\n    m = mokkari.api(username, password)\n\n    # Get all Marvel comics for the week of 2021-06-07\n    this_week = m.issues_list({"store_date_range_after": "2021-06-07", "store_date_range_before": "2021-06-13", "publisher": "marvel"})\n\n    # Print the results\n    for i in this_week:\n        print(f"{i.id} {i.issue_name}")\n\n    # Retrieve the detail for an individual issue\n    asm_68 = m.issue(31660)\n\n    # Print the issue Description\n    print(asm_68.desc)\n  \nDocumentation\n-------------\n- `Read the project documentation <https://mokkari.readthedocs.io/en/latest/>`_\n\nBugs/Requests\n-------------\n  \nPlease use the `GitHub issue tracker <https://github.com/bpepple/mokkari/issues>`_ to submit bugs or request features.\n\nLicense\n-------\n\nThis project is licensed under the `GPLv3 License <LICENSE>`_.\n',
    'author': 'Brian Pepple',
    'author_email': 'bdpepple@gmail.com',
    'maintainer': 'Brian Pepple',
    'maintainer_email': 'bdpepple@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
PublishersList module.

This module provides the following classes:

- PublishersList
"""
from marshmallow import ValidationError

from mokkari import exceptions, publisher


class PublishersList:
    """The PublishersList object contains a list of `Publisher` objects."""

    def __init__(self, response):
        """Initialize a new PublishersList."""
        self.publishers = []

        schema = publisher.PublisherSchema()
        for pub_dict in response["results"]:
            try:
                result = schema.load(pub_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.publishers.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.publishers)

    def __len__(self):
        """Return the length of the object."""
        return len(self.publishers)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.publishers[index]



"""
SQLite Cache module.

This module provides the following classes:

- SqliteCache
"""
import json
import sqlite3


class SqliteCache:
    """
    The SqliteCache object to cache search results from Metron.

    :param str db_name: Path and database name to use.
    """

    def __init__(self, db_name: str = "mokkari_cache.db"):
        """Intialize a new SqliteCache."""
        self.con = sqlite3.connect(db_name)
        self.cur = self.con.cursor()
        self.cur.execute("CREATE TABLE IF NOT EXISTS responses (key, json)")

    def get(self, key: str):
        """
        Retrieve data from the cache database.

        :param str key: value to search for.
        """
        self.cur.execute("SELECT json FROM responses WHERE key = ?", (key,))
        result = self.cur.fetchone()

        if result:
            return json.loads(result[0])

        return None

    def store(self, key: str, value: str):
        """
        Save data to the cache database.

        :param str key: Item id.
        :param str value: data to save.
        """
        self.cur.execute(
            "INSERT INTO responses(key, json) VALUES(?, ?)", (key, json.dumps(value))
        )
        self.con.commit()



"""
IssuesList module.

This module provides the following classes:

- IssuesList
"""
from marshmallow import ValidationError

from mokkari import exceptions, issue


class IssuesList:
    """The IssuesList object contains a list of `Issue` objects."""

    def __init__(self, response):
        """Initialize a new IssuesList."""
        self.issues = []

        schema = issue.IssueSchema()
        for issue_dict in response["results"]:
            try:
                result = schema.load(issue_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.issues.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.issues)

    def __len__(self):
        """Return the length of the object."""
        return len(self.issues)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.issues[index]



"""
Test Issues module.

This module contains tests for Issue objects.
"""
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal

import pytest
import requests_mock

from mokkari import exceptions, issues_list


def test_known_issue(talker):
    """Test for a known issue."""
    death = talker.issue(1)
    assert death.publisher.name == "Marvel"
    assert death.series.name == "Death of the Inhumans"
    assert death.volume == 1
    assert death.story_titles[0] == "Chapter One: Vox"
    assert death.cover_date == date(2018, 9, 1)
    assert death.store_date == date(2018, 7, 4)
    assert death.price is None
    assert not death.sku
    assert death.image == "https://static.metron.cloud/media/issue/2018/11/11/6497376-01.jpg"
    assert len(death.characters) > 0
    assert len(death.teams) > 0
    assert len(death.credits) > 0
    assert death.modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        16,
        899872,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )
    assert death.teams[0].name == "Inhumans"
    assert death.teams[0].id == 1
    assert death.teams[0].modified == datetime(
        2019,
        6,
        23,
        15,
        13,
        23,
        975156,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )


def test_issue_with_price_and_sku(talker):
    """Test issue with price & sku values."""
    die_16 = talker.issue(36860)
    assert die_16.price == Decimal("3.99")
    assert die_16.sku == "JUN210207"
    assert die_16.cover_date == date(2021, 8, 1)
    assert die_16.store_date == date(2021, 8, 25)


def test_issue_without_store_date(talker):
    """Test issue that does not have a store date."""
    spidey = talker.issue(31047)
    assert spidey.publisher.name == "Marvel"
    assert spidey.series.name == "The Spectacular Spider-Man"
    assert spidey.volume == 1
    assert spidey.story_titles[0] == "A Night on the Prowl!"
    assert spidey.cover_date == date(1980, 10, 1)
    assert spidey.store_date is None
    assert "Dennis O'Neil" in [c.creator for c in spidey.credits]
    assert "Spider-Man" in [c.name for c in spidey.characters]


def test_issue_without_story_title(talker):
    """Test an issue that does not have a story title."""
    redemption = talker.issue(30662)
    assert redemption.publisher.name == "AWA Studios"
    assert redemption.series.name == "Redemption"
    assert redemption.volume == 1
    assert len(redemption.story_titles) == 0
    assert redemption.cover_date == date(2021, 5, 1)
    assert redemption.store_date == date(2021, 5, 19)
    assert "Christa Faust" in [c.creator for c in redemption.credits]


def test_issueslist(talker):
    """Test the IssueList."""
    issues = talker.issues_list({"series_name": "action comics", "series_year_began": 2011})
    issue_iter = iter(issues)
    assert next(issue_iter).id == 6730
    assert next(issue_iter).id == 6731
    assert next(issue_iter).id == 6732
    assert len(issues) == 57
    assert issues[2].id == 6732
    assert issues[56].issue_name == "Action Comics #52"


def test_issueslist_with_params(talker):
    """Test the IssueList with params given."""
    params = {
        "series_name": "Kang",
    }
    issues = talker.issues_list(params=params)
    assert len(issues) == 2
    assert issues[0].issue_name == "Kang The Conqueror #1"
    assert issues[0].cover_date == date(2021, 10, 1)


def test_issue_with_upc_sku_price(talker):
    """Test issue with upc, sku, and price values."""
    usca_3 = talker.issue(36812)
    assert usca_3.series.name == "The United States of Captain America"
    assert usca_3.number == "3"
    assert usca_3.price == Decimal("4.99")
    assert usca_3.sku == "JUN210696"
    assert usca_3.upc == "75960620100600311"


def test_issue_without_upc_sku_price(talker):
    """Test issue without upc, sku, and price values."""
    bullets = talker.issue(32662)
    assert bullets.price is None
    assert bullets.sku == ""
    assert bullets.upc == ""


def test_issue_with_variants(talker):
    """Test issue with variant data."""
    paprika = talker.issue(37094)
    assert paprika.series.id == 2511
    assert paprika.series.name == "Mirka Andolfo's Sweet Paprika"
    assert paprika.volume == 1
    assert paprika.number == "2"
    assert paprika.cover_date == date(2021, 9, 1)
    assert paprika.store_date == date(2021, 9, 1)
    assert paprika.price == Decimal("3.99")
    assert paprika.sku == "JUN210256"
    assert paprika.page_count is None
    assert len(paprika.credits) == 9
    assert len(paprika.variants) == 4
    assert paprika.variants[0].name == "Cover B Sejic"
    assert paprika.variants[0].sku == "JUN210257"
    assert (
        paprika.variants[0].image
        == "https://static.metron.cloud/media/variants/2021/08/26/sweet-paprika-2b.jpg"
    )
    assert paprika.variants[1].name == "Cover C March"
    assert paprika.variants[1].sku == "JUN210258"
    assert (
        paprika.variants[1].image
        == "https://static.metron.cloud/media/variants/2021/08/26/sweet-paprika-2c.jpg"
    )


def test_issue_with_page_count(talker):
    """Test issue that has a page count."""
    gr = talker.issue(8118)
    assert gr.page_count == 40
    assert gr.number == "1"
    assert gr.upc == "75960609672500111"
    assert gr.cover_date == date(2020, 2, 1)
    assert gr.store_date == date(2019, 12, 18)
    assert gr.series.name == "Revenge of the Cosmic Ghost Rider"
    assert gr.volume == 1


def test_bad_issue(talker):
    """Test for a non-existant issue."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/issue/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.issue(-1)


def test_bad_response_data():
    """Test for bad issue response."""
    with pytest.raises(exceptions.ApiError):
        issues_list.IssuesList({"results": {"volume": "1"}})


def test_multi_page_results(talker):
    """Test for multi page results."""
    issues = talker.issues_list({"series_name": "action comics", "series_year_began": 1938})
    assert len(issues) == 864
    assert issues[0].issue_name == "Action Comics #1"
    assert issues[0].cover_date == date(1938, 6, 1)
    assert issues[863].issue_name == "Action Comics #904"
    assert issues[863].cover_date == date(2011, 10, 1)



"""
Exceptions module.

This module provides the following classes:

- ApiError
- AuthenticationError
- CacheError
"""


class ApiError(Exception):
    """Class for any api errors."""

    def __init__(self, *args, **kwargs):
        """Initialize an ApiError."""
        Exception.__init__(self, *args, **kwargs)


class AuthenticationError(ApiError):
    """Class for any authentication errors."""

    def __init__(self, *args, **kwargs):
        """Initialize an AuthenticationError."""
        Exception.__init__(self, *args, **kwargs)


class CacheError(ApiError):
    """Class for any database cache errors."""

    def __init__(self, *args, **kwargs):
        """Initialize an CacheError."""
        Exception.__init__(self, *args, **kwargs)



"""
CharactersList module.

This module provides the following classes:

- CharactersList
"""
from marshmallow import ValidationError

from mokkari import character, exceptions


class CharactersList:
    """The CharactersList object contains a list of `Character` objects."""

    def __init__(self, response):
        """Initialize a new CharactersList."""
        self.characters = []

        schema = character.CharacterSchema()
        for character_dict in response["results"]:
            try:
                result = schema.load(character_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.characters.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.characters)

    def __len__(self):
        """Return the length of the object."""
        return len(self.characters)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.characters[index]



"""
Issue module.

This module provides the following classes:

- Role
- RolesSchema
- Credit
- CreditsSchema
- Issue
- IssueSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load

from mokkari import arc, character, publisher, series, team, variant


class Role:
    """
    The Role object contains information for creators' role.

    :param `**kwargs`: The keyword arguments is used for setting role data from Metron.
    """

    def __init__(self, **kwargs):
        """Intialize a new Role."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class RolesSchema(Schema):
    """Schema for the Roles."""

    id = fields.Int()
    name = fields.Str()

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Role object.

        :param data: Data from Metron response.

        :returns: :class:`Role` object
        :rtype: Role
        """
        return Role(**data)


class Credit:
    """
    The Credit object contains information for creators credits for an issue.

    :param `**kwargs`: The keyword arguments is used for setting creator credit data.
    """

    def __init__(self, **kwargs):
        """Intialize a new Credit."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CreditsSchema(Schema):
    """Schema for the Credits."""

    id = fields.Int()
    creator = fields.Str()
    role = fields.Nested(RolesSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Credit object.

        :param data: Data from Metron response.

        :returns: :class:`Credit` object
        :rtype: Credit
        """
        return Credit(**data)


class Issue:
    """
    The Issue object contains information for an issue.

    :param `**kwargs`: The keyword arguments is used for setting issue data from Metron.
    """

    def __init__(self, **kwargs):
        """Intialize a new Issue."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class IssueSchema(Schema):
    """
    Schema for the Issue API.

    .. versionchanged:: 0.1.6

        - ``name`` field  changed to ``story_titles``
        - ``__str__`` field change to ``issue_name``

    .. versionchanged:: 0.2.0
        Added ``price`` and ``sku`` fields

    .. versionchanged:: 0.2.2
        Added ``upc`` field

    .. versionchanged:: 0.2.4

        - Added ``page_count`` field
        - Changed ``price`` field from a string to float value.

    .. versionchanged:: 1.0.0

        - Changed ``price`` field to a decimal type.
        - Added ``modified`` field

    """

    id = fields.Int()
    publisher = fields.Nested(publisher.PublisherSchema)
    series = fields.Nested(series.SeriesSchema)
    volume = fields.Int()
    number = fields.Str()
    story_titles = fields.List(fields.Str(allow_none=True), data_key="name")
    cover_date = fields.Date()
    store_date = fields.Date(allow_none=True)
    price = fields.Decimal(places=2, allow_none=True)
    sku = fields.Str()
    upc = fields.Str()
    page_count = fields.Int(allow_none=True, data_key="page")
    desc = fields.Str(allow_none=True)
    image = fields.URL()
    arcs = fields.Nested(arc.ArcSchema, many=True)
    credits = fields.Nested(CreditsSchema, many=True)
    characters = fields.Nested(character.CharacterSchema, many=True)
    teams = fields.Nested(team.TeamSchema, many=True)
    issue_name = fields.Str(data_key="__str__")
    variants = fields.Nested(variant.VariantSchema, many=True)
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the arc object.

        :param data: Data from Metron response.

        :returns: :class:`Issue` object
        :rtype: Issue
        """
        return Issue(**data)
